#include <assert.h>
#include <string.h>
#include <tlssep-config.h>
#include <libconfig.h>

static config_t config, *ptr  = NULL;
static const char *application_path = NULL;

typedef enum {
	LOOKUP_ARRAY,
	LOOKUP_BOOLEAN,
	LOOKUP_INT,
	LOOKUP_STRING,
} lookup_type_t;

bool
tlssep_config_init(const char *config_path, const char *_application_path)
{
	bool fnval = true;

	application_path = _application_path;

	config_init(&config);
	ptr = &config;
	assert(NULL != ptr);

	if (!config_read_file(&config, config_path)) {
		fprintf(stderr, "%s:%d - %s\n", config_error_file(&config), config_error_line(&config), config_error_text(&config));
		fnval = false;
		goto done;
	}

done:
	if (false == fnval && NULL != ptr) {
		config_destroy(&config);
	}

	return fnval;
}

static bool
_lookup(lookup_type_t type, const char *name, const void **value)
{
	assert(NULL != ptr);
	assert(NULL != application_path);
	assert(NULL != name);
	assert(NULL != value);

	bool fnval = false;
	static char option_name[PATH_MAX + 128 + 1]; // This makes 128 max option name length.

	assert(strlen(name) <= 128);

	strcpy(option_name, application_path);
	strcat(option_name, ".");
	strcat(option_name, name);

	config_setting_t *root = config_root_setting(ptr);
	if (NULL == root) {
		goto done;
	}

	config_setting_t *programs = config_setting_get_member(root, "programs");
	if (NULL != programs) {
		int i = 0;
		config_setting_t *group;
		for (i = 0, group = config_setting_get_elem(programs, i); NULL != group; i++, group = config_setting_get_elem(programs, i)) {
			const char *path = NULL;
			fnval = config_setting_lookup_string(group, "path", &path);
			if (CONFIG_TRUE != fnval) {
				continue;
			}
			if (strcmp(path, application_path)) {
				continue;
			}
			switch (type) {
			case LOOKUP_ARRAY:
				*value = config_lookup_from(group, name);
				if (NULL != *value) {
					fnval = true;
					goto done;
				}
				break;
			case LOOKUP_BOOLEAN:
				fnval = config_setting_lookup_bool(group, name, (int *) value);
				if (CONFIG_TRUE == fnval) {
					goto done;
				}
				break;
			case LOOKUP_INT:
				fnval = config_setting_lookup_int(group, name, (int *) value);
				if (CONFIG_TRUE == fnval) {
					goto done;
				}
				break;
			case LOOKUP_STRING:
				fnval = config_setting_lookup_string(group, name, (const char **) value);
				if (CONFIG_TRUE == fnval) {
					goto done;
				}
				break;
			default:
				goto done;
			}
		}
	}

	switch (type) {
	case LOOKUP_ARRAY:
		*value = config_lookup(ptr, name);
		if (NULL != *value) {
			fnval = true;
		}
		break;
	case LOOKUP_BOOLEAN:
		fnval = config_lookup_bool(ptr, name, (int *) value);
		break;
	case LOOKUP_INT:
		fnval = config_lookup_int(ptr, name, (int *) value);
		break;
	case LOOKUP_STRING:
		fnval = config_lookup_string(ptr, name, (const char **) value);
		break;
	default:
		goto done;
	}

done:
	return fnval;
}

bool
tlssep_config_lookup_certificate_path(const char **cert_path)
{
	return _lookup(LOOKUP_STRING, "certpath", (const void **) cert_path);
}

bool
tlssep_config_lookup_privkey_path(const char **privkey_path)
{
	return _lookup(LOOKUP_STRING, "privkeypath", (const void **) privkey_path);
}

bool
tlssep_config_lookup_enough(int *enough)
{
	return _lookup(LOOKUP_INT, "enough", (const void **) enough);
}

bool
tlssep_config_lookup_require_client_certificate(bool *require_client_certificate)
{
	int value;
	bool fnval = _lookup(LOOKUP_BOOLEAN, "require-client-certificate", (const void **) &value);
	*require_client_certificate = value != 0;
	return fnval;
}

bool
tlssep_config_lookup_provide_client_certificate(bool *provide_client_certificate)
{
	int value;
	bool fnval = _lookup(LOOKUP_BOOLEAN, "provide-client-certificate", (const void **) &value);
	*provide_client_certificate = value != 0;
	return fnval;
}

bool
tlssep_config_lookup_verification_engines(int *engine_mask)
{
	assert(NULL != ptr);

	*engine_mask = 0;

	config_setting_t *engines;
	bool fnval = _lookup(LOOKUP_ARRAY, "verification", (const void **) &engines);
	if (false == fnval) {
		goto done;
	}

	int i;
	const char *engine_str;
	for (i = 0, engine_str = config_setting_get_string_elem(engines, i); NULL != engine_str; i++, engine_str = config_setting_get_string_elem(engines, i)) {
		if (0 == strcmp("unsafe", engine_str)) {
			*engine_mask |= TLSSEP_VERIFICATION_ENGINE_UNSAFE;
		} else if (0 == strcmp("ca", engine_str)) {
			*engine_mask |= TLSSEP_VERIFICATION_ENGINE_CA;
		} else if (0 == strcmp("self-signed", engine_str)) {
			*engine_mask |= TLSSEP_VERIFICATION_ENGINE_SELF_SIGNED;
		} else {
			fnval = false;
			goto done;
		}
	}

done:
	return fnval;
}

void
tlssep_config_free(void)
{
	assert(NULL != ptr);
	config_destroy(ptr);
}
