#include <openssl/x509v3.h>
#include <assert.h>
#include <tlssep-verify.h>

static tlssep_status_t
_verify_self_signed(SSL *ssl, X509 *cert)
{
	X509_STORE_CTX *ctx = NULL;
	X509_STORE *store = NULL;
	tlssep_status_t fnval = TLSSEP_STATUS_OK;

	if (NULL == cert) {
		fnval = TLSSEP_STATUS_ERROR_DECORATOR_MISSING_CERTIFICATE;
		goto done;
	}

	/* Based on work from Zakir Durumeric at
	 * https://zakird.com/2013/10/13/certificate-parsing-with-openssl/.
	 * This seems more robust than merely checking for and accepting:
	 * 	X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
	 * 	or
	 * 	X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN
	 * from SSL_get_verify_result, as SSL_get_verify_result
	 * notes that it can only return a single error even
	 * when multiple errors exist.
	 *
	 * TODO: Do these allocations unnecessarily affect performance?
	 */

	store = X509_STORE_new();
	if (NULL == store) {
		fnval = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		goto done;
	}

	STACK_OF(X509) *sk = SSL_get_peer_cert_chain(ssl);
	if (NULL == sk) {
		fnval = TLSSEP_STATUS_ERROR_DECORATOR_MISSING_CERTIFICATE_CHAIN;
		goto done;
	}

	int num = sk_X509_num(sk);
	if (0 != num) {
		X509 *top = sk_X509_value(sk, num - 1);
		if (num != 0 && NULL == top) {
			fnval = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
			goto done;
		}

		X509_STORE_add_cert(store, top);
	}

	/* SSL_get_peer_cert_chain does not return end cert when called on
	 * server side! Add it here just to be sure. */
	X509_STORE_add_cert(store, cert);

	ctx = X509_STORE_CTX_new();
	if (NULL == ctx) {
		fnval = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		goto done;
	}

	X509_STORE_CTX_init(ctx, store, cert, sk);

	int result = X509_verify_cert(ctx);
	if (1 != result) {
		fnval = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE;
		goto done;
	}

done:
	if (NULL != ctx) {
		X509_STORE_CTX_free(ctx);
	}

	if (NULL != store) {
		X509_STORE_free(store);
	}

	return fnval;
}

static tlssep_status_t
_verify_ca(SSL *ssl, X509 *cert)
{
	tlssep_status_t fnval = TLSSEP_STATUS_OK;

	if (NULL == cert) {
		fnval = TLSSEP_STATUS_ERROR_DECORATOR_MISSING_CERTIFICATE;
		goto done;
	}

	long result = SSL_get_verify_result(ssl);
	if (X509_V_OK != result) {
		fnval = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE;
		goto done;
	}

done:

	return fnval;
}

static tlssep_status_t
_extract_common_name(X509 *cert, char *expected_common_name, char *common_name) {
	assert(NULL != cert);
	assert(NULL != expected_common_name);

	static char common_name_scratch[TLSSEP_MAX_COMMON_NAME_SIZE];
	tlssep_status_t fnval = TLSSEP_STATUS_ERROR_DECORATOR_CERTIFICATE_MISMATCH;
	GENERAL_NAMES *alt_names = NULL;

	if (NULL == common_name) {
		common_name = common_name_scratch;
	}

	common_name[0] = 0x00;

	alt_names = X509_get_ext_d2i (cert, NID_subject_alt_name, NULL, NULL);
	if (NULL != alt_names) {
		int i, count = sk_GENERAL_NAME_num (alt_names);
		for (i = 0; i < count; i++) {
			GENERAL_NAME *name = sk_GENERAL_NAME_value (alt_names, i);
			if (NULL != name) {
				if (GEN_DNS == name->type) {
					unsigned char *name_utf8;
					ASN1_STRING_to_UTF8 (&name_utf8, name->d.dNSName);
					if (strlen ((char *) name_utf8) != ASN1_STRING_length (name->d.dNSName)) {
						/* Register www.google.com[0x00]example.com?
						 * See http://www.blackhat.com/presentations/bh-usa-09/MARLINSPIKE/BHUSA09-Marlinspike-DefeatSSL-PAPER1.pdf
						 */
						fnval = TLSSEP_STATUS_ERROR_DECORATOR_CERTIFICATE_NAME_CONTAINS_NULL;
						goto done;
					}
					bool ok = ! strcmp(expected_common_name, (char *) name_utf8);
					OPENSSL_free (name_utf8);
					if (true == ok) {
						strcpy(common_name, expected_common_name);
						fnval = TLSSEP_STATUS_OK;
						goto done;
					}
				}
			}
		}
	}

	X509_NAME *subj = X509_get_subject_name(cert);
	if (NULL == subj) {
		goto done;
	}

	X509_NAME_get_text_by_NID (subj, NID_commonName, common_name, TLSSEP_MAX_COMMON_NAME_SIZE);

	int i = X509_NAME_get_index_by_NID(subj, NID_commonName, 0);
	if (-1 == i) {
		goto done;
	}

	X509_NAME_ENTRY *entry = X509_NAME_get_entry(subj, i);
	if (NULL == entry) {
		goto done;
	}

	ASN1_STRING *data = X509_NAME_ENTRY_get_data(entry);
	if (NULL == data) {
		goto done;
	}

	if (strlen (common_name) != (size_t) ASN1_STRING_length (data)) {
		/* Register www.google.com[0x00]example.com?
		 * See http://www.blackhat.com/presentations/bh-usa-09/MARLINSPIKE/BHUSA09-Marlinspike-DefeatSSL-PAPER1.pdf
		 */
		fnval = TLSSEP_STATUS_ERROR_DECORATOR_CERTIFICATE_NAME_CONTAINS_NULL;
		goto done;
	}

	if (0 == strcmp(expected_common_name, common_name)) {
		fnval = TLSSEP_STATUS_OK;
	}

done:
	return fnval;
}

tlssep_status_t
tlssep_certificate_ok(SSL *ssl, int engine_mask, int enough, char *expected_common_name, char *common_name)
{
	assert(NULL != ssl);
	assert(NULL != expected_common_name); /* "" means no hostname check. */

	tlssep_status_t cn_match_status = TLSSEP_STATUS_ERROR;
	int verification_count = 0;
	tlssep_status_t fnval = TLSSEP_STATUS_OK;

	X509 *cert = SSL_get_peer_certificate(ssl);
	if (NULL != cert) {
		cn_match_status = _extract_common_name(cert, expected_common_name, common_name);
	}

	int i;
	for (i = 0; i < TLSSEP_VERIFICATION_ENGINE_LAST; i++) {
		int engine = 1 << i;
		if (engine_mask & engine) {
			switch (engine) {
			case TLSSEP_VERIFICATION_ENGINE_UNSAFE:
				fnval = TLSSEP_STATUS_OK;
				break;
			case TLSSEP_VERIFICATION_ENGINE_CA:
				fnval = _verify_ca(ssl, cert);
				break;
			case TLSSEP_VERIFICATION_ENGINE_SELF_SIGNED:
				fnval = _verify_self_signed(ssl, cert);
				break;
			default:
				fnval = TLSSEP_STATUS_ERROR_DECORATOR_BAD_VERIFICATION_ENGINE;
				goto done;
			}

			if (TLSSEP_STATUS_OK != fnval) {
				continue;
			}

			if (TLSSEP_VERIFICATION_ENGINE_UNSAFE != engine) {
				if (TLSSEP_STATUS_OK != cn_match_status) {
					fnval = cn_match_status;
					continue;
				}
			}

			verification_count++;
		}
	}

done:
	if (NULL != cert) {
		X509_free(cert);
	}

	if (TLSSEP_STATUS_OK == fnval) {
		fnval = verification_count >= enough ? TLSSEP_STATUS_OK : TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE;
	} else {
		fnval = verification_count >= enough ? TLSSEP_STATUS_OK : fnval;
	}

	return fnval;
}
