#ifndef __TLSSEP_H
#define __TLSSEP_H

#include <sys/param.h>
#include <stdbool.h>
#include <tlssep-status.h>

#define TLSSEP_MAX_COMMON_NAME_SIZE 64

typedef unsigned int _tlssep_id_t;

/**
 * tlssep_context_t:
 *
 * Maintains the state of a set of TLS connections, but contains no public fields.
 * Initialize this structure by calling #tlssep_init.
 */
typedef struct tlssep_context_t {
	/*< private >*/
	pid_t   decoratorPid;
	int     opSocket;
	int     fdTransferSocket;
} tlssep_context_t;

/**
 * tlssep_desc_t:
 *
 * Serves as the descriptor of a single TLS connection. An application can
 * poll the @notificationSocket file descriptor to determine if there exists
 * data which can be read from the network decorator using #tlssep_read.
 * Initialize this structure by calling #tlssep_connect or #tlssep_accept.
 */
typedef struct tlssep_desc_t {
	int               notificationSocket;
	/*< private >*/
	tlssep_context_t *context;
	_tlssep_id_t       id;
} tlssep_desc_t;

/**
 * tlssep_init:
 * @context: a pointer to a #tlssep_context_t which this function will initialize 
 *
 * This function initializes libtlssep and executes a network decorator
 * in preparation for later creating a TLS session.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_init        (tlssep_context_t *context);

/**
 * tlssep_connect:
 * @context: a pointer to a valid #tlssep_context_t
 * @file_descriptor: the file descriptor upon which to create a TLS session; usually a network socket
 * @expected_common_name: a pointer to NULL or a valid string representing the common name which is expected to be present in the remote certificate; what is done with this depends on the configured verification engine
 * @common_name: if not NULL, this function will copy the remote certificate's common name to this address; the buffer must be at least @TLSSEP_MAX_COMMON_NAME_SIZE bytes long
 * @desc: a pointer to a #tlssep_desc_t which this function will initialize
 *
 * This function creates a TLS session by connecting over the given file descriptor.
 * Most commonly, the file descriptor is a socket with an existing transport-layer connection to the remote host.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_connect     (tlssep_context_t *context, int file_descriptor, const char *expected_common_name, char *common_name, tlssep_desc_t *desc);

/**
 * tlssep_accept:
 * @context: a pointer to a valid #tlssep_context_t
 * @file_descriptor: the file descriptor upon which to create a TLS session; usually a network socket
 * @expected_common_name: a pointer to NULL or a valid string representing the common name which is expected to be present in the remote certificate; what is done with this depends on the configured verification engine
 * @common_name: if not NULL, this function will copy the remote certificate's common name to this address; the buffer must be at least @TLSSEP_MAX_COMMON_NAME_SIZE bytes 
 * long
 * @desc: a pointer to a #tlssep_desc_t which this function will initialize
 *
 * This function creates a TLS session by accepting on the given file descriptor.
 * Most commonly, the file descriptor is a socket with an existing transport-layer connection to the remote host.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_accept      (tlssep_context_t *context, int file_descriptor, const char *expected_common_name, char *common_name, tlssep_desc_t *desc);

/**
 * tlssep_write:
 * @desc: a pointer to a valid #tlssep_desc_t
 * @buf: some buffer of bytes to write
 * @buf_size: the number of bytes to write from @buf
 * @bytes_written: a pointer to an integer which this function will set to the number of bytes actually written
 *
 * This function passes a number of bytes to the network decorator which will encrypt the bytes before writing them to the TLS connection.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_write       (tlssep_desc_t *desc, const void *buf, int buf_size, int *bytes_written);

/**
 * tlssep_read:
 * @desc: a pointer to a valid #tlssep_desc_t
 * @buf: a buffer which will hold any bytes read
 * @buf_size: the size of @buf in bytes
 * @bytes_read: a pointer to an integer which this function will set to the number of bytes actually read
 *
 * This function requests that a number of bytes will be read by the network decorator.
 * The network decorator will decrypt the bytes read from the TLS connection before returning them to the application.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_read        (tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read);

/**
 * tlssep_peek:
 * @desc: a pointer to a valid #tlssep_desc_t
 * @buf: a buffer which will hold any bytes read
 * @buf_size: the size of @buf in bytes
 * @bytes_read: a pointer to an integer which this function will set to the number of bytes actually read
 *
 * This function requests that a number of bytes will be read by the network decorator.
 * The network decorator will decrypt the bytes read from the TLS connection before returning them to the application. Unlike with #tlssep_read, the network decorator will not remove the bytes from its TLS buffer; thus subsequent #tlssep_read/#tlssep_peek calls will read the same bytes again.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_peek        (tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read);

/**
 * tlssep_poll:
 * @desc: a pointer to a valid #tlssep_desc_t
 * @timeout: the number of seconds to wait for data before giving up
 *
 * This function blocks until there is data ready to be read from the TLS connection
 * up to the limit of @timeout seconds.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_poll        (tlssep_desc_t *desc, unsigned int timeout);

/**
 * tlssep_setnonblock:
 * @desc: a pointer to a valid #tlssep_desc_t
 *
 * This function sets the mode of the network decorator’s TLS file
 * descriptor to non-blocking.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_setnonblock (tlssep_desc_t *desc);

/**
 * tlssep_close:
 * @desc: a pointer to a valid #tlssep_desc_t
 *
 * This function instructs the decorator to close the given TLS
 * connection and remove its file descriptor from the select file descriptor set. The
 * procedure also frees any state associated with the connection
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_close       (tlssep_desc_t *desc);

/**
 * tlssep_terminate:
 * @context: a pointer to a valid #tlssep_context_t
 *
 * This function instructs the network decorator to exit.
 *
 * Returns: a #tlssep_status_t
 */
tlssep_status_t tlssep_terminate   (tlssep_context_t *context);

/**
 * tlssep_strerror:
 * @error: a #tlssep_status_t
 *
 * This function transforms a #tlssep_status_t into its string description.
 *
 * Returns: the string description of the given #tlssep_status_t
 */
char           *tlssep_strerror    (tlssep_status_t error);

#endif
